/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
/* gdome-xml-nnm.c
 *
 * Copyright (C) 1999 Raph Levien <raph@acm.org>
 * Copyright (C) 2000 Mathieu Lacage <mathieu@gnu.org>
 * Copyright (C) 2000 Anders Carlsson <andersca@gnu.org>
 * Copyright (C) 2001 Paolo Casarini <paolo@casarini.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <glib.h>
#include "gdome.h"
#include "gdome-util.h"
#include <tree.h>
#include <parser.h>
#include <xmlmemory.h>
#include <hash.h>
#include "gdome-xml-util.h"
#include "gdome-xml-document.h"
#include "gdome-xml-element.h"
#include "gdome-xml-nnodem.h"
#include "gdome-xml-xmlutil.h"

typedef struct _nodeIterator nodeIterator;
struct _nodeIterator {
	unsigned long cur;
	unsigned long index;
	xmlNode *node;
};

const GdomeNamedNodeMapVtab gdome_xml_nnm_vtab = {
	gdome_xml_nnm_ref,
	gdome_xml_nnm_unref,
	gdome_xml_nnm_query_interface,
	gdome_xml_nnm_getNamedItem,
	gdome_xml_nnm_setNamedItem,
	gdome_xml_nnm_removeNamedItem,
	gdome_xml_nnm_item,
	gdome_xml_nnm_length
};

void *
itemHashScanner (void *payload, void *data, xmlChar *name) {
  nodeIterator *priv = (nodeIterator *)data;

	if(priv->cur < priv->index)
    priv->cur++;
  else if(priv->node == NULL)
    priv->node = (xmlNode *)payload;

  return NULL;
}

/**
 * gdome_xml_nnm_mkref:
 * @doc  the Document owner ref
 * @elem  the Element owner ref, can be NULL
 * @data  the pointer to the structure to export as NamedNodeMap
 * @readonly  TRUE if the GdomeNamedNodeMap is readonly
 * @type  the type of the <code>data</code> structure passed
 *
 * Construct the NamedNodeMap Object.
 * Returns a pointer to the GdomeNamedNodeMap Object.
 */
GdomeNamedNodeMap *
gdome_xml_nnm_mkref (GdomeDocument *doc, GdomeElement *elem, void *data, GdomeAccessType accessType, gint type) {
	Gdome_xml_NamedNodeMap *result = NULL;

	if (doc == NULL)
    return NULL;

	result = g_new (Gdome_xml_NamedNodeMap, 1);
	result->refcnt = 1;
	result->super.vtab = &gdome_xml_nnm_vtab;

	result->doc = doc;
  result->elem = elem;
	if(type == GDOME_XML_NNM_HASH)
		result->data.ht = (xmlHashTable *)data;
	else if(type == GDOME_XML_NNM_LIST)
		result->data.lst = (xmlNode *)data;
	else {
		g_free(result);
		return NULL;
	}

	result->accessType = accessType;
	result->type = type;

	return (GdomeNamedNodeMap *) result;
}


void
gdome_xml_nnm_ref (GdomeNamedNodeMap *self, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	
	if (self == NULL)
		return;
	if (exc == NULL)
		return;

	priv->refcnt++;
}

void
gdome_xml_nnm_unref (GdomeNamedNodeMap *self, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;

	if (priv == NULL)
		return ;
	if (exc == NULL)
		return ;

	priv->refcnt--;

	if (priv->refcnt == 0)
		g_free (self);
}

void *
gdome_xml_nnm_query_interface (GdomeNamedNodeMap *self, const char *interface, GdomeException *exc) {
	return NULL;
}

/**
 * gdome_xml_nnm_length:
 * @self:  NamedNodeMap Objects ref
 * @exc:  Exception Object ref
 *
 * Returns the number of nodes in this map. The range of valid child node
 * indices is 0 to length-1 inclusive.
 */
unsigned long
gdome_xml_nnm_length (GdomeNamedNodeMap *self, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	xmlHashTable *ht;
	xmlNode *n;
	long result = 0L;

	if(priv == NULL)
		return 0L;
	if(exc == NULL)
		return 0L;

  /* HashTable */
	if (priv->type == GDOME_XML_NNM_HASH) {
		ht = priv->data.ht;

    if((result = xmlHashSize(ht)) == -1)
      return 0L;
	}
  /* List */
	else if (priv->type == GDOME_XML_NNM_LIST) {
		for(n = priv->data.lst; n != NULL; n = n->next, result++);
	}
	
	return result;
}

/**
 * gdome_xml_nnm_getNamedItem:
 * @self:  NamedNodeMap Objects ref
 * @name:  The nodeName of the node to retrieve.
 * @exc:  Exception Object ref
 *
 * Retrieves a node specified by name.
 * Returns a Node (of any type) with the specified nodeName, or null if it
 * does not identify any node in this map.
 */
GdomeNode *
gdome_xml_nnm_getNamedItem (GdomeNamedNodeMap *self, GdomeDOMString *name, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	xmlHashTable *ht = NULL;
	xmlNode *n = NULL;

	if (priv == NULL)
		return NULL;
	if (name == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;

  /* HashTable */
	if (priv->type == GDOME_XML_NNM_HASH) {
		ht = priv->data.ht;

    n = (xmlNode*) xmlHashLookup (ht, (xmlChar *) name->str);
	}
  /* List */
  else if (priv->type == GDOME_XML_NNM_LIST)
    n = (xmlNode *)gdome_xmlGetAttr(((Gdome_xml_Element *)priv->elem)->n,
                                    (xmlChar *)name->str);

	if(n != NULL)
		return (GdomeNode *) gdome_xml_n_mkref (n);
	else
		return NULL;
}

/**
 * gdome_xml_nnm_getNamedItemNS:
 * @self:  NamedNodeMap Objects ref
 * @namespaceURI:  The namespace URI of the node to retrieve.
 * @localName:  The local name of the node to retrieve.
 * @exc:  Exception Object ref
 *
 * Retrieves a node specified by local name and namespace URI.
 * Returns a Node (of any type) with the specified local name and namespace
 * URI, or null if they do not identify any node in this map.
 */
GdomeNode *
gdome_xml_nnm_getNamedItemNS (GdomeNamedNodeMap *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	xmlHashTable *ht = NULL;
	xmlNode *n = NULL;

	if (priv == NULL)
		return NULL;
	if (namespaceURI == NULL)
		return NULL;
	if (localName == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;

  /* HashTable */
	if (priv->type == GDOME_XML_NNM_HASH) {
		ht = priv->data.ht;

    n = (xmlNode*) xmlHashLookup2 (ht, (xmlChar *) localName->str,
                                   (xmlChar *) namespaceURI->str);
	}
  /* List */
	else if (priv->type == GDOME_XML_NNM_LIST)
    n = (xmlNode *)gdome_xmlGetNsAttr(((Gdome_xml_Element *)priv->elem)->n,
                                      (xmlChar *)namespaceURI->str,
                                      (xmlChar *)localName->str);

	if(n != NULL)
		return (GdomeNode *) gdome_xml_n_mkref (n);
	else
		return NULL;
}

/**
 * gdome_xml_nnm_item:
 * @self:  NamedNodeMap Objects ref
 * @index:  Index into this map.
 * @exc:  Exception Object ref
 *
 * Returns the indexth item in the map. If index is greater than or equal to
 * the number of nodes in this map, this returns null.
 *
 * Note: it computes its task in O(n).
 */
GdomeNode *
gdome_xml_nnm_item (GdomeNamedNodeMap *self, unsigned long index, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	xmlNode *n = NULL;
	xmlHashTable *ht;
	nodeIterator *iter;
  unsigned long cur;

	if(priv == NULL)
		return NULL;
	if(exc == NULL)
		return NULL;

  /* HashTable */
	if (priv->type == GDOME_XML_NNM_HASH) {
		ht = priv->data.ht;
		iter = g_new(nodeIterator, 1);

    iter->cur = 0L;
		iter->index = index;
    iter->node = NULL;
		xmlHashScan(ht, itemHashScanner, iter);
		n = iter->node;

    g_free(iter);
	}
  /* List */
	else if (priv->type == GDOME_XML_NNM_LIST)
    for(n = priv->data.lst, cur = 0;
        n != NULL && cur < index;
        cur++,	n = n->next);

	return gdome_xml_n_mkref (n);
}

/**
 * gdome_xml_nnm_removeNamedItem:
 * @self:  NamedNodeMap Objects ref
 * @name:  The nodeName of the node to remove.
 * @exc:  Exception Object ref
 *
 * Removes a node specified by name. When this map contains the attributes
 * attached to an element, if the removed attribute is known to have a default
 * value, an attribute immediately appears containing the default value as
 * well as the corresponding namespace URI, local name, and prefix when applicable.
 * Returns the node removed from this map if a node with such a name exists.
 *
 * GDOME_NOT_FOUND_ERR: Raised if there is no node named <code>name</code> in
 * this map.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
 */
GdomeNode *
gdome_xml_nnm_removeNamedItem (GdomeNamedNodeMap *self, GdomeDOMString *name, GdomeException *exc) {
  Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
  xmlNode *ret = NULL;

	if (priv == NULL)
		return NULL;
	if (name == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;
	if (priv->accessType == GDOME_READONLY_NODE) {
		*exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
		return NULL;
	}

  /* Hash Table */
	if (priv->type == GDOME_XML_NNM_HASH) {
		GdomeNode *gret = NULL;
    xmlHashTable *ht = priv->data.ht;

    if((gret = gdome_xml_nnm_getNamedItem(self, name, exc)) != NULL) {
      xmlHashRemoveEntry(ht, name->str, NULL);
			return gret;
    }
	}
  /* List */
  else if (priv->type == GDOME_XML_NNM_LIST) {
    xmlNode *parent = ((Gdome_xml_Element *)priv->elem)->n;
    xmlChar *defaultvalue = NULL;

    ret = (xmlNode *)gdome_xmlUnlinkAttr(parent, name->str);
		if(ret != NULL) {
      /* resync structure */
			priv->data.lst = (xmlNode *)gdome_xmlGetAttrList(parent);

      /* check for a default value and eventually insert it */
      defaultvalue = xmlGetProp(parent, name->str);
			if(defaultvalue != NULL) {
         xmlNewProp(parent, name->str, defaultvalue);
         xmlFree(defaultvalue);
			}

      /* Check for ID removal */
			if (xmlIsID(parent->doc, parent, (xmlAttr *)ret))
				xmlRemoveID(parent->doc, (xmlAttr *)ret);

			return (GdomeNode *) gdome_xml_n_mkref (ret);
		}
	}

	*exc = GDOME_NOT_FOUND_ERR;
	return NULL;
}

/**
 * gdome_xml_nnm_removeNamedItemNS:
 * @self:  NamedNodeMap Objects ref
 * @namespaceURI:  The namespace URI of the node to remove.
 * @localName:  The local name of the node to remove.
 * @exc:  Exception Object ref
 *
 * Removes a node specified by local name and namespace URI. When this map
 * contains the attributes attached to an element, if the removed attribute
 * is known to have a default value, an attribute immediately appears
 * containing the default value as well as the corresponding namespace URI,
 * local name, and prefix when applicable.
 * Returns the node removed from this map if a node with such a local name and
 * namespace URI exists.
 *
 * GDOME_NOT_FOUND_ERR: Raised if there is no node named <code>name</code> in
 * this map.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
 */
GdomeNode *
gdome_xml_nnm_removeNamedItemNS (GdomeNamedNodeMap *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;

	if (priv == NULL)
		return NULL;
	if (namespaceURI == NULL)
		return NULL;
	if (localName == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;
	if (priv->accessType == GDOME_READONLY_NODE) {
		*exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
		return NULL;
	}

  /* Hash Table */
	if (priv->type == GDOME_XML_NNM_HASH) {
    GdomeNode *ret = NULL;
    xmlHashTable *ht = priv->data.ht;

    if((ret = gdome_xml_nnm_getNamedItemNS(self, namespaceURI, localName, exc)) != NULL) {
      xmlHashRemoveEntry2(ht, namespaceURI->str, localName->str, NULL);
      return ret;
    }
	}
  /* List */
	else if (priv->type == GDOME_XML_NNM_LIST) {
    xmlNode *ret = NULL;
    xmlNode *parent = ((Gdome_xml_Element *)priv->elem)->n;
    xmlChar *defaultvalue = NULL;

		if((ret = (xmlNode *)gdome_xmlUnlinkNsAttr(parent, namespaceURI->str, localName->str)) != NULL) {
      /* resync structure */
			priv->data.lst = (xmlNode *)parent->properties;

      /* check for a default value and eventually insert it */
      defaultvalue = xmlGetNsProp(parent, localName->str, namespaceURI->str);
			if(defaultvalue != NULL) {
         xmlNewNsProp(parent, gdome_xmlGetNs(ret), localName->str, defaultvalue);
			}
      xmlFree(defaultvalue);

			/* Check for ID removal */
			if (xmlIsID(parent->doc, parent, (xmlAttr *)ret))
				xmlRemoveID(parent->doc, (xmlAttr *)ret);

			return (GdomeNode *) gdome_xml_n_mkref (ret);
		}
	}

	*exc = GDOME_NOT_FOUND_ERR;
	return NULL;
}

/**
 * gdome_xml_nnm_setNamedItem:
 * @self:  NamedNodeMap Objects ref
 * @arg:  a node to store in this map.
 * @exc:  Exception Object ref
 *
 * Adds a node using its nodeName attribute. If a node with that name is
 * already present in this map, it is replaced by the new one.
 * Returns If the new Node replaces an existing node the replaced Node is
 * returned, otherwise null is returned.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if arg was created from a different
 * document than the one that created this map.
 * GDOME NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
 * GDOME INUSE_ATTRIBUTE_ERR: Raised if arg is an Attr that is already an
 * attribute of another Element object. The DOM user must explicitly clone
 * Attr nodes to re-use them in other elements.
 */
GdomeNode *
gdome_xml_nnm_setNamedItem (GdomeNamedNodeMap *self, GdomeNode *arg, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	xmlNode *newn = NULL;
	xmlNode *ret = NULL;

	if (priv == NULL)
		return NULL;
	if (arg == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;
	if (((Gdome_xml_Document *)priv->doc)->n != ((Gdome_xml_Node*)arg)->n->doc) {
		*exc = GDOME_WRONG_DOCUMENT_ERR;
    return NULL;
	}
	if (priv->accessType == GDOME_READONLY_NODE) {
		*exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
		return NULL;
	}

	newn = ((Gdome_xml_Node*)arg)->n;

  /* Hash Table */
	if(priv->type == GDOME_XML_NNM_HASH) {
		xmlHashTable *ht = priv->data.ht;

		if((ret = xmlHashLookup(ht, gdome_xmlGetName(newn))) != NULL) {
			xmlHashUpdateEntry(ht, gdome_xmlGetName(newn), newn, NULL);
      return (GdomeNode *) gdome_xml_n_mkref(ret);
		} else {
			xmlHashAddEntry(ht, gdome_xmlGetName(newn), newn);
			return NULL;
		}
	}
  /* List */
	else if (priv->type == GDOME_XML_NNM_LIST) {
    xmlNode *parent = ((Gdome_xml_Element *)priv->elem)->n;

		if(!GDOME_XML_IS_A(arg))
			/* FIXME: we need a specific Gdome Exc for args error */
			return NULL;
    if(newn->prev || newn->next ||
       (gdome_xmlGetParent(newn) && gdome_xmlGetParent(newn) != parent)) {
      *exc = GDOME_INUSE_ATTRIBUTE_ERR;
			return NULL;
		}

    ret = (xmlNode *)gdome_xmlUnlinkAttr(parent, gdome_xmlGetName(newn));
    gdome_xmlLinkAttr(parent, (xmlAttr *)newn);

    /* resync structure */
    priv->data.lst = (xmlNode *)gdome_xmlGetAttrList(parent);

		if(ret != NULL)
			return (GdomeNode *) gdome_xml_n_mkref(ret);
	}

	return NULL;
}

/**
 * gdome_xml_nnm_setNamedItemNS:
 * @self:  NamedNodeMap Objects ref
 * @arg:  a node to store in this map. The node will later be accessible
 *        using the value of its namespaceURI and localName attributes.
 * @exc:  Exception Object ref
 *
 * Adds a node using its namespaceURI and localName. If a node with that
 * namespace URI and that local name is already present in this map, it is
 * replaced by the new one
 * Returns If the new Node replaces an existing node the replaced Node is
 * returned, otherwise null is returned.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if arg was created from a different
 * document than the one that created this map.
 * NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
 * INUSE_ATTRIBUTE_ERR: Raised if arg is an Attr that is already an attribute
 * of another Element object. The DOM user must explicitly clone Attr nodes to
 * re-use them in other elements.
 */
GdomeNode *
gdome_xml_nnm_setNamedItemNS (GdomeNamedNodeMap *self, GdomeNode *arg, GdomeException *exc) {
	Gdome_xml_NamedNodeMap *priv = (Gdome_xml_NamedNodeMap *)self;
	xmlNode *newn = NULL;
	xmlNode *ret = NULL;

	if (priv == NULL)
		return NULL;
	if (arg == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;
	if (((Gdome_xml_Document *)priv->doc)->n != ((Gdome_xml_Node*)arg)->n->doc) {
		*exc = GDOME_WRONG_DOCUMENT_ERR;
    return NULL;
	}
	if (priv->accessType == GDOME_READONLY_NODE) {
		*exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
		return NULL;
	}

	newn = ((Gdome_xml_Node*)arg)->n;

  /* Hash Table */
	if(priv->type == GDOME_XML_NNM_HASH) {
		xmlHashTable *ht = priv->data.ht;

		if((ret = xmlHashLookup2(ht, gdome_xmlGetName(newn), gdome_xmlGetNsURI(newn))) != NULL) {
			xmlHashUpdateEntry2(ht, gdome_xmlGetName(newn), gdome_xmlGetNsURI(newn), newn, NULL);
      return (GdomeNode *) gdome_xml_n_mkref(ret);
		} else {
			xmlHashAddEntry2(ht, gdome_xmlGetName(newn), gdome_xmlGetNsURI(newn), newn);
			return NULL;
		}

	}
  /* List */
	else if (priv->type == GDOME_XML_NNM_LIST) {
    xmlNode *parent = ((Gdome_xml_Element *)priv->elem)->n;

		if(!GDOME_XML_IS_A(arg) || gdome_xmlGetNs(newn) == NULL)
			/* FIXME: we need a specific Gdome Exc for args error */
			return NULL;
    if(newn->prev || newn->next ||
       (gdome_xmlGetParent(newn) && gdome_xmlGetParent(newn) != parent)) {
      *exc = GDOME_INUSE_ATTRIBUTE_ERR;
			return NULL;
		}

    ret = (xmlNode *)gdome_xmlUnlinkNsAttr(parent, gdome_xmlGetNsURI(newn), gdome_xmlGetName(newn));
    gdome_xmlLinkAttr(parent, (xmlAttr *)newn);

    /* resync structure */
    priv->data.lst = (xmlNode *)gdome_xmlGetAttrList(parent);

		if(ret != NULL)
			return (GdomeNode *) gdome_xml_n_mkref(ret);
	}
	return NULL;
}
